@@ -2,41 +2,90 @@
  * Copyright (c) 1989, 1990, 1991 by the University of Washington
  *
  * For copying and distribution information, please see the file
- * <uw-copyright.h>.
+ * <copyright.h>.
  *
- * Xarchie status calls added by George Ferguson, 13 Aug 1991.
+ * v1.2.4 - 11/19/91 (mmt) - added MSDOS & OS2 stuff
+ * v1.2.3 - 11/04/91 (bcn) - removed host comparison and replaced with check
+ *			     for connection id (undoes effect of v1.2.2.).
+ * v1.2.2 - 11/02/91 (gf)  - removed extra inet_ntoa() calls and stuff for
+ *			     multi-interface nets (lmjm@doc.imperial.ac.uk)
+ * v1.2.1 - 10/20/91 (gf)  - asynch implementation
+ * v1.2.0 - 09/17/91 (bpk) - added BULL & USG stuff, thanks to Jim Sillas
+ * v1.1.2 - 08/30/91 (bpk) - added VMS support
+ * v1.1.1 - 08/29/91 (bcn) - changed backoff handling
+ * v1.1.0 - 08/13/91 (gf)  - added XArchie status calls
+ *
+ * gf: 20 Oct 1991:
+ *  Broken into pieces so that under X dirsend() doesn't block in select()
+ *  but rather uses Xt calls to allow continued event processing. If
+ *  XARCHIE is not defined, can still be used since processEvent() will
+ *  use select() in this case.
  */
 
-#include <uw-copyright.h>
+#include <copyright.h>
 #include <stdio.h>
 #include <errno.h>
-#include <sys/time.h>
-#include <netdb.h>
-#include <strings.h>
 
-#include <sys/types.h>
-#include <sys/socket.h>
-#ifdef AIX
-#include <sys/select.h>
-#endif
+#ifdef VMS
+# ifdef WALLONGONG
+#  include "twg$tcp:[netdist.include]netdb.h"
+# else /* Multinet */
+#  include "multinet_root:[multinet.include]netdb.h"
+# endif
+# include <vms.h>
+#else /* not VMS */
+# include <sys/types.h> /* this may/will define FD_SET etc */
+# include <pmachine.h>
+# ifdef NEED_TIME_H
+#  include <time.h>
+# else
+#  include <sys/time.h>
+# endif
+# ifdef NEED_STRING_H
+#  include <string.h>
+# else
+#  include <strings.h>
+# endif
+# include <netdb.h>
+# include <sys/socket.h>
+# ifdef NEED_SELECT_H
+#  include <sys/select.h>
+# endif /* NEED_SELECT_H */
+# ifndef IN_H
+#  include <netinet/in.h>
+#  define IN_H
+# endif
+# ifndef hpux
+#  include <arpa/inet.h>
+# endif
+#endif /* !VMS */
 
 #include <pfs.h>
 #include <pprot.h>
 #include <pcompat.h>
 #include <perrno.h>
-#include <pmachine.h>
 
+/* Gnu C currently fails to pass structures on Sparcs properly.  This directly
+   effects the calling of inet_ntoa().  To get around it, we use this hack;
+   take the address of what's being called to inet_ntoa, so it gets it
+   properly.  This won't be necessary with gcc 2.0.  */
+#if defined(sun) && defined(__GNUC__) && !defined(__GNU_LIBRARY__)
+# define SUN_GNU_FIX &
+#else
+# define SUN_GNU_FIX
+#endif
 
 static int notprived = 0;
+#ifndef MSDOS
 extern int errno;
+#endif
 extern int perrno;
+extern int rdgram_priority;
 #ifdef DEBUG
 extern int pfs_debug;
 #endif
 extern int pfs_disable_flag;
 
-int	rdgram_priority = 0;
-
 char	*nlsindex();
 
 #define max(X, Y)  ((X) > (Y) ? (X) : (Y))
@@ -53,6 +102,93 @@ static int client_dirsrv_timeout = CLIENT_DIRSRV_TIMEOUT;
 static int client_dirsrv_retry = CLIENT_DIRSRV_RETRY; 
 #endif /* XARCHIE */
 
+/* These were parameters to dirsend() */
+static PTEXT pkt;
+static char *hostname;
+static struct sockaddr_in *hostaddr;
+
+/* These were locals in dirsend(). Note that the initializations here
+ * are really meaningless since we have to redo them for each call to
+ * dirsend() since they were formerly automatically initialized.
+ */
+static PTEXT		first = NULL;	/* First returned packet	 */
+static PTEXT		next;		/* The one we are waiting for 	 */
+static PTEXT		vtmp;           /* For reorganizing linked list  */
+static PTEXT		comp_thru;	/* We have all packets though    */
+static int		lp = -1;	/* Opened UDP port	         */
+static int		hdr_len;	/* Header Length                 */
+static int		nd_pkts;	/* Number of packets we want     */
+static int		no_pkts;	/* Number of packets we have     */
+static int		pkt_cid;        /* Packet connection identifier  */
+static unsigned short	this_conn_id;	/* Connection ID we are using    */
+static unsigned short	recvd_thru;	/* Received through              */
+static short		priority;	/* Priority for request          */
+static short		one = 0;	/* Pointer to value 1            */
+static short		zero = 0;	/* Pointer to value 0		 */
+static char		*seqtxt;	/* Pointer to text w/ sequence # */
+static struct sockaddr_in  us;		/* Our address                   */
+static struct sockaddr_in  to;		/* Address to send query	 */
+static struct sockaddr_in  from;	/* Reply received from		 */
+static int		from_sz;	/* Size of from structure	 */
+static struct hostent	*host;		/* Host info from gethostbyname  */
+static long		newhostaddr;    /* New host address from *host   */
+static int		req_udp_port=0; /* Requested port (optional)     */
+static char		*openparen;	/* Delimits port in name         */
+static char		hostnoport[500];/* Host name without port        */
+static int		ns;		/* Number of bytes actually sent */
+static int		nr;		/* Number of bytes received      */
+static fd_set		readfds;	/* Used for select		 */
+static int		tmp;
+static char		*ctlptr;	/* Pointer to control field      */
+static short		stmp;		/* Temp short for conversions    */
+static int		backoff;	/* Server requested backoff      */
+static unsigned char	rdflag11;	/* First byte of flags (bit vect)*/
+static unsigned char	rdflag12;	/* Second byte of flags (int)    */
+static int		scpflag = 0;	/* Set if any sequencd cont pkts */
+static int		ackpend = 0;    /* Acknowledgement pending      */
+static int		gaps = 0;	/* Gaps present in recvd pkts   */
+static struct timeval	timeout;	/* Time to wait for response    */
+static struct timeval	ackwait;	/* Time to wait before acking   */
+static struct timeval	gapwait;	/* Time to wait b4 filling gaps */
+static struct timeval	*selwait;	/* Time to wait for select      */
+static int		retries;	/* was = client_dirsrv_retry    */
+char   to_hostname[512];		/* lmjm: saves inet_ntoa() str  */
+
+/* These are added so dirsend() "blocks" properly */
+static PTEXT dirsendReturn;
+static int dirsendDone;
+
+/* And here are the values for dirsendDone */
+#define DSRET_DONE		1
+#define DSRET_SEND_ERROR	-1
+#define DSRET_RECV_ERROR	-2
+#define DSRET_SELECT_ERROR	-3
+#define DSRET_TIMEOUT		-4
+#define DSRET_ABORTED		-5
+
+/* New procedures to break up dirsend() */
+static int initDirsend();
+static void retryDirsend(), keepWaitingDirsend();
+static void timeoutProc();
+static void readProc();
+
+/* Wrappers around X calls to allow non-X usage */
+static void addInputSource(), removeInputSource();
+static void addTimeOut(), removeTimeOut();
+static void processEvent();
+
+/* Extra stuff for the asynchronous X version of dirsend() */
+#ifdef XARCHIE
+#include <X11/Intrinsic.h>
+extern XtAppContext appContext;
+#else
+typedef char *XtPointer;
+typedef char *XtInputId;
+typedef char *XtIntervalId;
+#endif
+static XtInputId inputId;
+static XtIntervalId timerId = (XtIntervalId)0;
+
 /*
  * dirsend - send packet and receive response
  *
@@ -73,54 +209,65 @@ static int client_dirsrv_retry = CLIENT_DIRSRV_RETRY;
  *   The packet is freed even if dirsend fails.
  */
 PTEXT
-dirsend(pkt,hostname,hostaddr)
-    PTEXT pkt;
-    char *hostname;
-    struct sockaddr_in	*hostaddr;
+dirsend(pkt_p,hostname_p,hostaddr_p)
+    PTEXT pkt_p;
+    char *hostname_p;
+    struct sockaddr_in	*hostaddr_p;
 {
-    PTEXT		first = NULL;	/* First returned packet	 */
-    PTEXT		next;		/* The one we are waiting for 	 */
-    PTEXT		vtmp;           /* For reorganizing linked list  */
-    PTEXT		comp_thru;	/* We have all packets though    */
-    int			lp = -1;	/* Opened UDP port	         */
-    int			hdr_len;	/* Header Length                 */
-    int			nd_pkts;	/* Number of packets we want     */
-    int			no_pkts;	/* Number of packets we have     */
-    int			pkt_cid;        /* Packet connection identifier  */
-    unsigned short	this_conn_id;	/* Connection ID we are using    */
-    unsigned short	recvd_thru;	/* Received through              */
-    short		priority;	/* Priority for request          */
-    static short	one = 0;	/* Pointer to value 1            */
-    static short	zero = 0;	/* Pointer to value 0		 */
-    char		*seqtxt;	/* Pointer to text w/ sequence # */
-    struct sockaddr_in  us;		/* Our address                   */
-    struct sockaddr_in	to;		/* Address to send query	 */
-    struct sockaddr_in	from;		/* Reply received from		 */
-    int			from_sz;	/* Size of from structure	 */
-    struct hostent	*host;		/* Host info from gethostbyname  */
-    long		newhostaddr;    /* New host address from *host   */
-    int			req_udp_port=0; /* Requested port (optional)     */
-    char		*openparen;	/* Delimits port in name         */
-    char		hostnoport[500];/* Host name without port        */
-    int			ns;		/* Number of bytes actually sent */
-    int			nr;		/* Number of bytes received      */
-    fd_set		readfds;	/* Used for select		 */
-    int			tmp;
-    char		*ctlptr;	/* Pointer to control field      */
-    short		stmp;		/* Temp short for conversions    */
-    int			backoff;	/* Server requested backoff      */
-    unsigned char	rdflag11;	/* First byte of flags (bit vect)*/
-    unsigned char	rdflag12;	/* Second byte of flags (int)    */
-    int			scpflag = 0;	/* Set if any sequencd cont pkts */
+    /* copy parameters to globals since other routines use them */
+    pkt = pkt_p;
+    hostname = hostname_p;
+    hostaddr = hostaddr_p;
+    /* Do the initializations of formerly auto variables */
+    first = NULL;
+    lp = -1;
+    one = 0;
+    zero = 0;
+    req_udp_port=0;
+    scpflag = 0;
+    ackpend = 0;
+    gaps = 0;
+    retries = client_dirsrv_retry;
 
-    int			ackpend = 0;    /* Acknowledgement pending      */
-    int			gaps = 0;	/* Gaps present in recvd pkts   */
-    struct timeval	timeout;	/* Time to wait for response    */
-    struct timeval	ackwait;	/* Time to wait before acking   */
-    struct timeval	gapwait;	/* Time to wait b4 filling gaps */
-    struct timeval	*selwait;	/* Time to wait for select      */
-    int			retries = client_dirsrv_retry;
+    if (initDirsend() < 0)
+	return(NULL);
+    addInputSource();
+    /* set the first timeout */
+    retryDirsend();
 
+    dirsendReturn = NULL;
+    dirsendDone = 0;
+    /* Until one of the callbacks says to return, keep processing events */
+    while (!dirsendDone)
+	processEvent();
+    /* Clean up event generators */
+    removeTimeOut();
+    removeInputSource();
+#ifdef XARCHIE
+    /* Set status if needed (has to be outside of loop or X will crash) */
+    switch (dirsendDone) {
+	case DSRET_SEND_ERROR: status0("Send error"); break;
+	case DSRET_RECV_ERROR: status0("Recv error"); break;
+        case DSRET_TIMEOUT:
+		status1("Connection to %s timed out",to_hostname);
+		break;
+        case DSRET_ABORTED: status0("Aborted"); break;
+    }
+#endif
+    /* Return whatever we're supposed to */
+    return(dirsendReturn);
+}
+
+
+/*	-	-	-	-	-	-	-	-	*/
+/* This function does all the initialization that used to be done at the
+ * start of dirsend(), including opening the socket descriptor "lp". It
+ * returns the descriptor if successful, otherwise -1 to indicate that
+ * dirsend() should return NULL immediately.
+ */
+static int
+initDirsend()
+{
 #ifdef XARCHIE
     status0("Initializing");
 #endif
@@ -194,7 +341,8 @@ dirsend(pkt,hostname,hostaddr)
 #endif
 	    perrno = DIRSEND_BAD_HOSTNAME;
 	    ptlfree(pkt);
-            return(NULL);
+            /* return(NULL); */
+	    return(-1);
 	}
 	/* If a port is included, save it away */
 	if(openparen = index(hostname,'(')) {
@@ -217,7 +365,8 @@ dirsend(pkt,hostname,hostaddr)
 #endif
 		perrno = DIRSEND_BAD_HOSTNAME;
 		ptlfree(pkt);
-		return(NULL);
+		/* return(NULL); */
+		return(-1);
 	    }
 	    bzero((char *)&to, S_AD_SZ);
 	    to.sin_family = AF_INET;
@@ -233,6 +382,10 @@ dirsend(pkt,hostname,hostaddr)
 	}
     }
     else bcopy(hostaddr,&to, S_AD_SZ);
+    /* lmjm: Save away the hostname */
+    strncpy(to_hostname,
+	    inet_ntoa(SUN_GNU_FIX *(struct in_addr *)&to.sin_addr),
+	    sizeof(to_hostname)-1);
 
     if(req_udp_port) to.sin_port = htons(req_udp_port);
     else to.sin_port = dir_udp_port;
@@ -252,7 +405,8 @@ dirsend(pkt,hostname,hostaddr)
 #endif
 	perrno = DIRSEND_UDP_CANT;
 	ptlfree(pkt);
-        return(NULL);
+        /* return(NULL); */
+	return(-1);
     }
 
     /* Try to bind it to a privileged port - loop through candidate */
@@ -303,13 +457,13 @@ dirsend(pkt,hostname,hostaddr)
         if (to.sin_family == AF_INET) {
 	    if(req_udp_port) 
 		fprintf(stderr,"Sending message to %s+%d(%d)...",
-			inet_ntoa(*(struct in_addr *)&to.sin_addr), req_udp_port, ntohs(this_conn_id));
+			to_hostname, req_udp_port, ntohs(this_conn_id));
 	    else fprintf(stderr,"Sending message to %s(%d)...",
-			 inet_ntoa(*(struct in_addr *)&to.sin_addr), ntohs(this_conn_id));
+			 to_hostname, ntohs(this_conn_id));
 	}
 #else
         if (to.sin_family == AF_INET) 
-	    fprintf(stderr,"Sending message to %s...", inet_ntoa(*(struct in_addr *)&to.sin_addr));
+	    fprintf(stderr,"Sending message to %s...", to_hostname);
 #endif
         else
             fprintf(stderr,"Sending message...");
@@ -321,18 +475,28 @@ dirsend(pkt,hostname,hostaddr)
     next = first;
 
 #ifdef XARCHIE
-    status1("Connecting to %s",inet_ntoa(*(struct in_addr *)&to.sin_addr));
+    status1("Connecting to %s",to_hostname);
 #endif
+    return(lp);
+}
 
- retry:
-
+/*	-	-	-	-	-	-	-	-	*/
+/*
+ * This used to be a label to goto to retry the last packet. Now we resend
+ * the packet and call keepWaitingDirsend() to wait for a reply. (We
+ * call keepWaitingDirsend() because formerly the code dropped through
+ * the keep_waiting label.)
+ */
+static void
+retryDirsend()
+{
     gaps = ackpend = 0;
 
     ns = sendto(lp,(char *)(pkt->start), pkt->length, 0, (struct sockaddr *)&to, S_AD_SZ);
     if(ns != pkt->length) {
 #ifdef DEBUG
 	if (pfs_debug) {
-	    fprintf(stderr,"\nsent only %d/%d: ",ns, pkt->length);
+    fprintf(stderr,"\nsent only %d/%d: ",ns, pkt->length);
 	    perror("");
 	}
 #endif
@@ -340,19 +504,30 @@ dirsend(pkt,hostname,hostaddr)
 	perrno = DIRSEND_NOT_ALL_SENT;
 	ptlfree(first);
 	ptlfree(pkt);
-#ifdef XARCHIE
-        status0("Send Error");
-#endif
-        return(NULL);
+	/* return(NULL); */
+	dirsendReturn = NULL;
+	dirsendDone = DSRET_SEND_ERROR;
     }
 #ifdef DEBUG
     if (pfs_debug > 2) fprintf(stderr,"Sent.\n");
 #endif
+    keepWaitingDirsend();
+}
+
+/*	-	-	-	-	-	-	-	-	*/
+/*
+ * This used to be a label to goto to set the appropriate timeout value
+ * and blocked in select(). Now we set selwait and the fd_sets to the
+ * appropriate values, and in X register a new timeout, then return to
+ * allow event processing.
+ */
+static void
+keepWaitingDirsend()
+{
     /* We come back to this point (by a goto) if the packet */
     /* received is only part of the response, or if the     */
     /* response came from the wrong host		    */
 
- keep_waiting:	
 #ifdef DEBUG
     if (pfs_debug > 2) fprintf(stderr,"Waiting for reply...");
 #endif
@@ -363,11 +538,21 @@ dirsend(pkt,hostname,hostaddr)
     else if(gaps) selwait = &gapwait;
     else selwait = &timeout;
 
-    /* select - either recv is ready, or timeout */
-    /* see if timeout or error or wrong descriptor */
-    tmp = select(lp + 1, &readfds, (fd_set *)0, (fd_set *)0, selwait);
+    addTimeOut();
+}
 
-    if((tmp == 0) && (gaps || ackpend)) { /* Send acknowledgment */
+/*	-	-	-	-	-	-	-	-	*/
+/*
+ * This routine is called when a timeout occurs. It includes the code that
+ * was formerly used when select() returned 0 (indicating a timeout).
+ */
+/*ARGSUSED*/
+static void
+timeoutProc(client_data,id)
+XtPointer client_data;
+XtIntervalId *id;
+{
+    if (gaps || ackpend) { /* Send acknowledgment */
 	/* Acks are piggybacked on retries - If we have received */
 	/* an ack from the server, then the packet sent is only  */
 	/* an ack and the rest of the message will be empty      */
@@ -377,10 +562,11 @@ dirsend(pkt,hostname,hostaddr)
 		    (ackpend ? "requested" : "gaps"));
 	}	
 #endif
-	goto retry;
+	retryDirsend();
+	return;
     }
 
-    if((tmp == 0) && (retries-- > 0)) {
+    if (retries-- > 0) {
 	timeout.tv_sec = CLIENT_DIRSRV_BACKOFF(timeout.tv_sec);
 #ifdef DEBUG
 	if (pfs_debug > 2) {
@@ -391,28 +577,41 @@ dirsend(pkt,hostname,hostaddr)
 #ifdef XARCHIE
         status1("Timed out -- retrying (%d seconds)",timeout.tv_sec);
 #endif
-	goto retry;
+	retryDirsend();
+	return;
     }
 
-    if((tmp < 1) || !FD_ISSET(lp, &readfds)) {
 #ifdef DEBUG
-	if (pfs_debug) {
-	    fprintf(stderr, "select failed: readfds=%x ",
-                    readfds);
-            perror("");
-        }
-#endif
-	close(lp);
-	perrno = DIRSEND_SELECT_FAILED;
-	ptlfree(first);
-	ptlfree(pkt);
-#ifdef XARCHIE
-        status1("Couldn't connect to %s",
-		inet_ntoa(*(struct in_addr *)&to.sin_addr));
-#endif
-        return(NULL);
+    if (pfs_debug) {
+	fprintf(stderr, "select failed(timeoutProc): readfds=%x ",
+		readfds);
+	perror("");
     }
+#endif
+    close(lp);
+    perrno = DIRSEND_SELECT_FAILED;
+    ptlfree(first);
+    ptlfree(pkt);
+    /* return(NULL); */
+    dirsendReturn = NULL;
+    dirsendDone = DSRET_TIMEOUT;
+}
 
+/*	-	-	-	-	-	-	-	-	*/
+/*
+ * This function is called whenever there's something to read on the
+ * connection. It includes the code that was run when select() returned
+ * greater than 0 (indicating read ready).
+ */
+/*ARGSUSED*/
+static void
+readProc(client_data,source,id)
+XtPointer client_data;
+int *source;
+XtInputId *id;
+{
+    /* We got something to read, so clear the timer */
+    removeTimeOut();
 
     from_sz = sizeof(from);
     next->start = next->dat;
@@ -424,10 +623,10 @@ dirsend(pkt,hostname,hostaddr)
 	perrno = DIRSEND_BAD_RECV;
 	ptlfree(first);
 	ptlfree(pkt);
-#ifdef XARCHIE
-        status0("Recv Error");
-#endif
-        return(NULL);
+	/* return(NULL) */
+	dirsendReturn = NULL;
+	dirsendDone = DSRET_RECV_ERROR;
+        return;
     }
 
     next->length = nr;
@@ -435,25 +634,19 @@ dirsend(pkt,hostname,hostaddr)
     *(next->start + next->length) = NULL;
 
 #ifdef DEBUG
-    if (pfs_debug > 2) 
-        fprintf(stderr,"Received packet from %s\n", inet_ntoa(*(struct in_addr *)&from.sin_addr));
+    if (pfs_debug > 2)
+        fprintf(stderr,"Received packet from %s\n",
+		inet_ntoa(SUN_GNU_FIX *(struct in_addr *)&from.sin_addr));
 #endif
-    if (to.sin_addr.s_addr != from.sin_addr.s_addr) {
-#ifdef DEBUG
-	if (pfs_debug) {
-	    fprintf(stderr, "dirsend: received packet from wrong host! (%x)\n",
-		    from.sin_addr.s_addr);
-	    (void) fflush(stdout);
-	}
-#endif
-	goto keep_waiting;
-    }
 
 #ifdef XARCHIE
     {
-        static int toggle;
-        status0(toggle ? "Receiving +" : "Receiving *");
-        toggle = !toggle;
+        static int num = 2;
+	if (num == 2)
+	    status1("Connected to %s",to_hostname);
+	else
+	    status1("Receiving...%c",(num?'+':'*'));
+	num = !num;
     }
 #endif
 
@@ -468,6 +661,12 @@ dirsend(pkt,hostname,hostaddr)
 	    if(stmp) pkt_cid = ntohs(stmp);
 	    ctlptr += 2;
 	}
+	if(pkt_cid && this_conn_id && (pkt_cid != this_conn_id)) {
+	    /* The packet is not for us */
+	    /* goto keep_waiting; */
+	    keepWaitingDirsend();
+	    return;
+	}
 	if(hdr_len >= 5) {	/* Packet number */
 	    bcopy(ctlptr,&stmp,2);
 	    next->seq = ntohs(stmp);
@@ -506,6 +705,10 @@ dirsend(pkt,hostname,hostaddr)
 		    fprintf(stderr,"Backing off to %d seconds\n", backoff);
 #endif
 		timeout.tv_sec = backoff;
+		if ((backoff > 60) && (first == next) && (no_pkts == 0)) {
+		    /* Probably a long queue on the server - don't give up */
+		    retries = client_dirsrv_retry;
+		}
 	    }
 	    ctlptr += 2;
 	}
@@ -533,7 +736,11 @@ dirsend(pkt,hostname,hostaddr)
 	    bcopy(ctlptr,&rdflag12,1);
 	    ctlptr += 1;
 	}
-	if(next->seq == 0) goto keep_waiting;
+	if(next->seq == 0) {
+	    /* goto keep_waiting; */
+	    keepWaitingDirsend();
+	    return;
+	}
 	if(next->length >= 0) next->length -= hdr_len;
 	next->start += hdr_len;
 	goto done_old;
@@ -554,6 +761,12 @@ dirsend(pkt,hostname,hostaddr)
 	bcopy(ctlptr,&stmp,2);
 	if(stmp) pkt_cid = ntohs(stmp);
 	ctlptr += 2;
+	if(pkt_cid && this_conn_id && (pkt_cid != this_conn_id)) {
+	    /* The packet is not for us */
+	    /* goto keep_waiting; */
+	    keepWaitingDirsend();
+	    return;
+	}
 	/* Packet number */
 	if(ctlptr < (next->start + next->length)) {
 	    bcopy(ctlptr,&stmp,2);
@@ -582,7 +795,11 @@ dirsend(pkt,hostname,hostaddr)
 	    if(backoff) timeout.tv_sec = backoff;
 	    ctlptr += 2;
 	}
-	if(next->seq == 0) goto keep_waiting;
+	if(next->seq == 0) {
+	    /* goto keep_waiting; */
+	    keepWaitingDirsend();
+	    return;
+	}
 	goto done_old;
 
     }
@@ -615,7 +832,9 @@ dirsend(pkt,hostname,hostaddr)
 	else gaps++;
 	no_pkts = 1;
 	next = ptalloc();
-	goto keep_waiting;
+	/* goto keep_waiting; */
+	keepWaitingDirsend();
+	return;
     }
 	
     if(comp_thru && (next->seq <= comp_thru->seq))
@@ -677,7 +896,9 @@ ins_done:
 
     if ((nd_pkts == 0) || (no_pkts < nd_pkts)) {
 	next = ptalloc();
-	goto keep_waiting;
+	/* goto keep_waiting; */
+	keepWaitingDirsend();
+	return;
     }
 
  all_done:
@@ -686,7 +907,7 @@ ins_done:
 	if (pfs_debug > 2) {
 	    if (to.sin_family == AF_INET)
 		fprintf(stderr,"Acknowledging final packet to %s(%d)\n",
-			inet_ntoa(*(struct in_addr *)&to.sin_addr), ntohs(this_conn_id));
+			to_hostname, ntohs(this_conn_id));
             else
                 fprintf(stderr,"Acknowledging final packet\n");
 	    (void) fflush(stderr);
@@ -731,6 +952,104 @@ ins_done:
 	}
     }
 
-    return(first);
+    /* return(first); */
+    dirsendReturn = first;
+    dirsendDone = DSRET_DONE;
 
 }
+/*	-	-	-	-	-	-	-	-	*/
+/* These routines allow dirsend() to run with or without X by providing
+ * wrappers around the calls that handle the asynchronous communication.
+ * All parameters are passed using globals.
+ * Under X: The input sources and timeouts are set using Xt calls, and
+ *	    processEvent() just calls XtAppProcessEvent().
+ * Non-X: None of the input sources and timeouts are used, and
+ *	  processEvent() calls select() to handle both timeouts and the
+ *	  socket file descriptor. The return value of select() is used
+ *	  to determine which callback routine to call.
+ */
+
+static void
+addInputSource()
+{
+#ifdef XARCHIE
+    inputId = XtAppAddInput(appContext,lp,XtInputReadMask,readProc,NULL);
+#endif
+}
+
+static void
+removeInputSource()
+{
+#ifdef XARCHIE
+    XtRemoveInput(inputId);
+#endif
+}
+
+static void
+addTimeOut()
+{
+#ifdef XARCHIE
+    unsigned long timeoutLen = selwait->tv_sec*1000 + selwait->tv_usec/1000;
+
+    /* old timeout can still be there if we are being called after the
+     * file descriptor was read, so we remove it just to be sure. */
+    removeTimeOut();
+    timerId = XtAppAddTimeOut(appContext,timeoutLen,timeoutProc,NULL);
+#endif
+}
+
+static void
+removeTimeOut()
+{
+#ifdef XARCHIE
+    if (timerId != (XtIntervalId)0) {
+	XtRemoveTimeOut(timerId);
+	timerId = (XtIntervalId)0;
+    }
+#endif
+}
+
+static void
+processEvent()
+{
+#ifdef XARCHIE
+    XtAppProcessEvent(appContext,XtIMAll);
+#else
+    /* select - either recv is ready, or timeout */
+    /* see if timeout or error or wrong descriptor */
+    tmp = select(lp + 1, &readfds, (fd_set *)0, (fd_set *)0, selwait);
+    if (tmp == 0) {
+	timeoutProc(NULL,&timerId);
+    } else if ((tmp < 0) || !FD_ISSET(lp,&readfds)) {
+#ifdef DEBUG
+	if (pfs_debug) {
+	    fprintf(stderr, "select failed(processEvent): readfds=%x ",
+		    readfds);
+	    perror("");
+	}
+#endif
+	close(lp);
+	perrno = DIRSEND_SELECT_FAILED;
+	ptlfree(first);
+	ptlfree(pkt);
+	/* return(NULL); */
+	dirsendReturn = NULL;
+	dirsendDone = DSRET_SELECT_ERROR;
+    } else {
+	readProc(NULL,&lp,&inputId);
+    }
+#endif /* XARCHIE */
+}
+
+void
+abortDirsend()
+{
+    if (!dirsendDone) {
+	close(lp);
+	ptlfree(first);
+	ptlfree(pkt);
+	dirsendReturn = NULL;
+	dirsendDone = DSRET_ABORTED;
+    }
+    return;
+}
